#include "log/Log.h"
#include "io/Serial.h"
#include "std/stdio.h"
#include "thread/PIT.h"
#include "thread/Scheduler.h"
#include <stdarg.h>

static int level_mask = 15;
static int backend_mask = 3;

static bool log_level_enabled(LogLevel level)
{
    return level_mask & (1 << (int)level);
}

static bool log_backend_enabled(Backend backend)
{
    return backend_mask & (1 << (int)backend);
}

static void log_backend_serial(const char* function, LogLevel level, const char* message, va_list origin)
{
    va_list ap;
    va_copy(ap, origin);
    Serial::reset_color();
    if (Scheduler::current_task() && Scheduler::current_task()->id)
    {
        Task* current_task = Scheduler::current_task();
        printf("[%ld.%ld] (%s %ld) %s: ", PIT::ms_since_boot / 1000, PIT::ms_since_boot % 1000, current_task->name,
               current_task->id, function);
    }
    else { printf("[%ld.%ld] (kernel) %s: ", PIT::ms_since_boot / 1000, PIT::ms_since_boot % 1000, function); }
    switch (level)
    {
    case LogLevel::WARN: Serial::set_color(Color::Yellow); break;
    case LogLevel::ERROR: Serial::set_color(Color::Red); break;
    default: break;
    }
    vprintf(message, ap);
    Serial::reset_color();
    va_end(ap);
}

static void log_backend_console(const char* function, [[maybe_unused]] LogLevel level, const char* message,
                                va_list origin)
{
    va_list ap;
    va_copy(ap, origin);
    kprintf("%s: ", function);
    vkprintf(message, ap);
    va_end(ap);
}

void KernelLog::log(const char* function, LogLevel level, const char* message, ...)
{
    if (!log_level_enabled(level)) return;
    va_list ap;
    va_start(ap, message);
    if (log_backend_enabled(Backend::Serial)) log_backend_serial(function, level, message, ap);
    if (log_backend_enabled(Backend::Console)) log_backend_console(function, level, message, ap);
    va_end(ap);
}

void KernelLog::logln(const char* function, LogLevel level, const char* message, ...)
{
    if (!log_level_enabled(level)) return;
    va_list ap;
    va_start(ap, message);
    if (log_backend_enabled(Backend::Serial))
    {
        log_backend_serial(function, level, message, ap);
        printf("\n");
    }
    if (log_backend_enabled(Backend::Console))
    {
        log_backend_console(function, level, message, ap);
        kprintf("\n");
    }
    va_end(ap);
}

void KernelLog::toggle_log_level(LogLevel level)
{
    level_mask ^= (1 << (int)level);
}

void KernelLog::toggle_log_backend(Backend backend)
{
    backend_mask ^= (1 << (int)backend);
}

void KernelLog::enable_log_backend(Backend backend)
{
    backend_mask |= (1 << (int)backend);
}